"""
Implementation of the austrian operator ÖBB (Österreichische Bundesbahnen).
"""

import datetime
import time
from typing import Dict, Optional, Any, Tuple

from .. import Train, Station, ConnectingTrain
from ...exceptions import DataInvalidError
from ...utils.conversions import kmh_to_ms
from ...utils.data import StaticDataConnector, DynamicDataConnector, JSONDataConnector, some_or_default, ScheduledEvent, \
    Position

API_BASE_URL_RAILNET_REGIO = "railnet.oebb.at"


class _RailnetStaticConnector(StaticDataConnector, JSONDataConnector):
    def __init__(self):
        super().__init__(base_url=API_BASE_URL_RAILNET_REGIO)

    def refresh(self):
        self.store("trainInfo", self._get("/api/trainInfo"))


class _RailnetDynamicConnector(DynamicDataConnector, JSONDataConnector):
    def __init__(self):
        super().__init__(base_url=API_BASE_URL_RAILNET_REGIO)

    def refresh(self):
        self.store("combined", self._get("/assets/modules/fis/combined.json", params={"_time": time.time()}))


class RailnetRegio(Train):
    def __init__(self):
        super().__init__()
        self._static_data = _RailnetStaticConnector()
        self._dynamic_data = _RailnetDynamicConnector()

    def now(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(
            int(some_or_default(
                self._dynamic_data.load('combined', {}).get('operationalMessagesInfo', {}).get('time'), default=0
            ))
        )

    @property
    def type(self) -> str:
        return self._static_data.load('trainInfo', {}).get('trainType', None)

    @property
    def number(self) -> str:
        return self._static_data.load('trainInfo', {}).get('lineNumber', None)

    @property
    def id(self) -> None:
        # Not available in Railnet
        return None

    @property
    def stations(self) -> Dict[Any, Station]:
        def to_datetime(data: str, __future: bool = None) -> Optional[datetime.datetime]:
            if some_or_default(data, default=None) is None:
                return None
            # Now
            now = self.now()
            # Assume that the day is the same as the current day
            obj = datetime.datetime.strptime(data, "%H:%M").replace(year=now.year, month=now.month, day=now.day)

            if __future is not None:
                # We know that the event is in the future,
                # but obj is currently in the past
                if __future and obj < now:
                    obj += datetime.timedelta(days=1)
                # We know that the event is in the past,
                # but obj is currently in the future
                elif not __future and obj > now:
                    obj -= datetime.timedelta(days=1)
            return obj

        stations = {}
        distance = 0.0
        future = False
        for station_json in self._dynamic_data.load('combined', {}).get('stationList', []):
            # The first station has no distance
            if station_json.get('distance', "") != "":
                distance += float(some_or_default(station_json.get('distance'), default=0.0))

            connections = list([
                ConnectingTrain(
                    train_type=connection.get('type', None),
                    line_number=connection.get('line', None),
                    platform=ScheduledEvent(
                        scheduled=connection.get('track', {}).get('all', None),
                        actual=connection.get('track', {}).get('all', None)
                    ),
                    destination=connection.get('destination', {}).get('all', None),
                    departure=ScheduledEvent(
                        scheduled=to_datetime(connection.get('schedule', None), True),
                        actual=to_datetime(connection.get('actual', None), True)
                    )
                )
                for connection in self._dynamic_data.load('combined').get('connectionInfo', {}).get('connections', [])
                if self._dynamic_data.load('combined').get('connectionInfo', {}).get('station_id', -1)
                == station_json.get('id', -2)
            ])

            stations[station_json.get('id')] = Station(
                station_id=station_json.get('id'),
                name=station_json.get('name', {}).get('all', None),
                platform=ScheduledEvent(
                    station_json.get('track', {}).get('all', None),
                    station_json.get('track', {}).get('all', None)
                ),
                arrival=ScheduledEvent(
                    to_datetime(station_json.get('arrivalSchedule', None), self.delay > 0 if future else False),
                    to_datetime(station_json.get('arrivalForecast', None), future)
                ),
                departure=ScheduledEvent(
                    to_datetime(station_json.get('departureSchedule', None), self.delay > 0 if future else False),
                    to_datetime(station_json.get('departureForecast', None), future)
                ),
                distance=distance,
                connections=connections
            )

            future = future or station_json.get('id', -1) == self._dynamic_data.load('combined', {}).get(
                'currentStation', {}).get('id', None)
        return stations

    @property
    def origin(self) -> Station:
        return super(RailnetRegio, self).origin

    @property
    def current_station(self) -> Station:
        # Get the current station id
        station_id = self._dynamic_data.load('combined', {}).get('currentStation', {}).get('id', -1)
        # Get the station from the stations dict
        try:
            return self.stations[station_id]
        except AttributeError as e:
            raise DataInvalidError("No current station found") from e

    @property
    def destination(self) -> Station:
        return super(RailnetRegio, self).destination

    @property
    def speed(self) -> float:
        return kmh_to_ms(
            float(self._dynamic_data.load('combined', {}).get('operationalMessagesInfo', {}).get('speed', 0)))

    @property
    def distance(self) -> float:
        try:
            return self.stations[self._dynamic_data.load(
                'combined', {}).get('operationalMessagesInfo', {}).get('distanceStation', -1)
                   ].distance + float(
                self._dynamic_data.load('combined', {}).get('operationalMessagesInfo', {}).get('distance', 0)
            )
        except AttributeError as e:
            raise DataInvalidError("No base station found for distance") from e

    @property
    def position(self) -> Position:
        map_info = self._dynamic_data.load('combined', {}).get('mapInfo', {})
        return Position(
            latitude=(
                float(map_info.get('latitude')) if some_or_default(map_info.get('latitude')) is not None else None
            ),
            longitude=(
                float(map_info.get('longitude')) if some_or_default(map_info.get('longitude')) is not None else None
            )
        )

    @property
    def delay(self) -> float:
        return float(self._dynamic_data.load('combined', {}).get('operationalMessagesInfo', {}).get('delay', 0))



"""
Abstract base classes for trains
"""

import abc
import datetime
import time
from typing import Optional, Tuple, Dict, Union, Any, List, Callable

from .. import Vehicle
from ..exceptions import DataInvalidError, APIConnectionError, InitialConnectionError
from ..utils.conversions import coordinates_to_distance
from ..utils.data import StaticDataConnector, DynamicDataConnector, ScheduledEvent, Position


class Station(object):
    """
    A Station is a stop on the trip
    """

    __slots__ = ["_id", "_name", "_platform", "_arrival", "_departure", "_position", "_distance", "_connections"]

    def __init__(self, station_id: Any, name: str, platform: ScheduledEvent[str] = None,
                 arrival: ScheduledEvent[datetime.datetime] = None, departure: ScheduledEvent[datetime.datetime] = None,
                 position: Position = None, distance: float = None,
                 connections: List["ConnectingTrain"] = None):
        """
        Initialize a new :class:`Station`

        :param station_id: The ID of the station
        :type station_id: Any
        :param name: The name of the station
        :type name: str
        :param platform: The platform that the vehicle is arriving at
        :type platform: ScheduledEvent[str]
        :param arrival: The arrival time at this station
        :type arrival: ScheduledEvent[datetime.datetime]
        :param departure: The departure time from this station
        :type departure: ScheduledEvent[datetime.datetime]
        :param position: The geographic position of the station
        :type position: Position
        :param distance: The distance from the start to this station
        :type distance: float
        :param connections: The connecting services departing from this station
        :type connections: List[ConnectingTrain]
        """
        self._id = station_id
        self._name = name
        self._platform = platform
        self._arrival = arrival
        self._departure = departure
        self._position = position
        self._distance = distance
        self._connections = connections

    @property
    def id(self) -> Any:
        """
        The unique ID of the station

        :return: The ID of the station
        :rtype: Any
        """
        return self._id

    @property
    def name(self) -> str:
        """
        The name of the station

        :return: The name of the station
        :rtype: str
        """
        return self._name

    @property
    def platform(self) -> ScheduledEvent[str]:
        """
        The platform that the train is arriving at

        :return: The platform
        :rtype: ScheduledEvent[str]
        """
        return self._platform

    @property
    def arrival(self) -> ScheduledEvent[datetime.datetime]:
        """
        The arrival time at this station

        :return: The datetime object of the arrival time encased in a ScheduledEvent
        :rtype: ScheduledEvent[datetime.datetime]
        """
        return self._arrival

    @property
    def departure(self) -> ScheduledEvent[datetime.datetime]:
        """
        The departure time at this station

        :return: The datetime object of the departure time encased in a ScheduledEvent
        :rtype: ScheduledEvent[datetime.datetime]
        """
        return self._departure

    @property
    def connections(self) -> List["ConnectingTrain"]:
        """
        The connecting services departing from this station

        :return: A list of ConnectingTrain objects
        :rtype: List[ConnectingTrain]
        """
        return self._connections

    @property
    def distance(self) -> float:
        """
        The distance from the start to this station in meters

        :return: The distance
        :rtype: float
        """
        return self._distance

    @property
    def position(self) -> Position:
        """
        The geographic position of the station

        :return: The coordinates of the station
        :rtype: Position
        """
        return self._position

    def calculate_distance(self, other: Union["Station", Tuple[float, float], int, float]) -> Optional[float]:
        """
        Calculate the distance in meters between this station and something else
        Accepts a :class:`Station`, a tuple of (latitude, longitude) or an integer for the distance calculation

        :param other: The other station or position to calculate the distance to
        :type other: Union[Station, Tuple[float, float], int, float]
        :return: The distance in meters
        :rtype: Optional[float]
        """
        # Get the position of the other station or the distance of the other station from the start
        if isinstance(other, Station):
            other = (
                other.distance
                if other.distance is not None
                else (
                    other.position
                    if other.position is not None
                    else None
                )
            )
        # If there is not enough information to calculate the distance, return None
        if other is None:
            return None

        # Calculate the distance
        if isinstance(other, int) or isinstance(other, float):
            return self.distance - other if self.distance - other >= 0 else other - self.distance
        if isinstance(other, tuple):
            return coordinates_to_distance((self.position.latitude, self.position.longitude), other)
        return None


class Train(Vehicle, metaclass=abc.ABCMeta):
    """
    Interface specifying the attributes of a train
    """

    __slots__ = ["_static_data", "_dynamic_data", "_initialized"]

    def __init__(self):
        self._initialized = False
        self._static_data: StaticDataConnector = ...
        self._dynamic_data: DynamicDataConnector = ...

    def __enter__(self):
        self.init()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()

    def init(self) -> None:
        """
        Initialize the :class:`DataConnector` objects.

        Calls ``refresh`` on the :class:`StaticDataConnector`
        and ``start`` on the :class:`DynamicDataConnector` by default.

        :return: Nothing
        :rtype: None
        """
        if self._initialized:
            return
        try:
            self._dynamic_data.start()
            self._static_data.refresh()
            self._initialized = True
        except APIConnectionError as e:
            raise InitialConnectionError() from e

    def shutdown(self) -> None:
        """
        Safely close the :class:`DataConnector` objects of this train.

        :return: Nothing
        :rtype: None
        """
        self._dynamic_data.stop()

    @property
    def connected(self) -> bool:
        """
        Whether the train is connected to the API

        :return: Whether the train is connected to the API
        :rtype: bool
        """
        return self._dynamic_data.connected

    def now(self) -> datetime.datetime:
        """
        Get the current time from the train

        :return: The current time
        :rtype: datetime.datetime
        """
        return datetime.datetime.now()

    def calculate_distance(self, station: Station) -> float:
        """
        Calculate the distance in meters between the train and a station

        :param station: The station to calculate the distance to
        :type station: Station
        :return: The distance in meters
        :rtype: float
        """
        return station.calculate_distance(self.distance or self.position)

    @property
    @abc.abstractmethod
    def id(self) -> str:
        """
        The unique ID of this specific train

        :return: The ID
        :rtype: str
        """
        pass

    @property
    @abc.abstractmethod
    def type(self) -> str:
        """
        The abbreviated train type

        :return: The train type
        :rtype: str
        """
        pass

    @property
    @abc.abstractmethod
    def number(self) -> str:
        """
        The line number of this train

        :return: The line number
        :rtype: str
        """
        pass

    @property
    @abc.abstractmethod
    def stations(self) -> Dict[Any, Station]:
        """
        The stations that this train passes through returned as a dict with the station ID as the key

        :return: The stations
        :rtype: Dict[str, Station]
        """
        pass

    @property
    @abc.abstractmethod
    def origin(self) -> Station:
        """
        The station where this train started the current journey

        :return: The first station on this trip
        :rtype: Station
        """
        stations = list(self.stations.values())
        if len(stations) > 0:
            return stations[0]
        raise DataInvalidError("No origin station found")

    @property
    @abc.abstractmethod
    def current_station(self) -> Station:
        """
        The station where this train will arrive next or is currently at

        :return: The current station
        :rtype: Station
        """
        # Get the current station id
        # Get the station from the stations dict
        pass

    @property
    @abc.abstractmethod
    def destination(self) -> Station:
        """
        The station where this train will end the current journey

        :return: The last station on this trip
        :rtype: Station
        """
        stations = list(self.stations.values())
        if len(stations) > 0:
            return stations[-1]
        raise DataInvalidError("No destination station found")

    @property
    @abc.abstractmethod
    def speed(self) -> float:
        """
        The current speed of the train in meters / second

        :return: The speed of the train
        :rtype: float
        """
        pass

    @property
    @abc.abstractmethod
    def distance(self) -> float:
        """
        The distance from the start in meters

        :return: The distance
        :rtype: float
        """
        pass

    @property
    @abc.abstractmethod
    def position(self) -> Position:
        """
        The current position of the train as a :class:`Position`

        :return: The position of the train
        :rtype: Position
        """
        pass

    @property
    @abc.abstractmethod
    def delay(self) -> float:
        """
        The current delay of the train in seconds

        :return: The delay of the train
        :rtype: float
        """
        return (self.current_station.arrival.actual - self.current_station.arrival.scheduled).total_seconds()


class ConnectingTrain(object):
    """
    A connecting train is a train that is not part of the main trip but of a connecting service

    It may only have limited information available
    """

    __slots__ = ["train_type", "line_number", "platform", "destination", "departure"]

    def __init__(self, train_type: Optional[str] = None, line_number: Optional[str] = None,
                 platform: Optional[ScheduledEvent[str]] = None, destination: Optional[str] = None,
                 departure: Optional[ScheduledEvent[datetime.datetime]] = None):
        self.train_type: Optional[str] = train_type
        """
        The abbreviated train type
        """
        self.line_number: Optional[str] = line_number
        """
        The line number of the train
        """
        self.platform: Optional[ScheduledEvent[str]] = platform
        """
        The platform where the train will depart from
        """
        self.departure: Optional[ScheduledEvent[datetime.datetime]] = departure
        """
        The departure time of the train
        """
        self.destination: Optional[str] = destination
        """
        The destination of the train
        """

    def __str__(self):
        return f"{self.train_type}{self.line_number} to {self.destination} ({self.departure}, platform {self.platform})"

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.train_type}{self.line_number} " \
               f"-> {self.destination} ({repr(self.departure)}, {repr(self.platform)})>"


class _LazyStation(Station):
    """
    The LazyStation is a Station that maybe does not yet have information on connecting trains.
    If it does not have information by the time it is requested by the user,
    it will then proceed to load the information through ``lazy_func``.
    """

    __slots__ = ["_lazy_func", "_cache_valid_until", "_cache_timeout"]

    def __init__(self, *args, lazy_func: Optional[Callable[..., List[ConnectingTrain]]] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._lazy_func = lazy_func
        self._cache_valid_until = 0
        self._cache_timeout = 60

    @property
    def connections(self) -> Optional[List[ConnectingTrain]]:
        def request_data():
            connections = self._lazy_func()
            self._connections = connections
            self._cache_valid_until = time.time() + self._cache_timeout  # Cache this result for the next minute
            return connections

        if self._lazy_func is None:
            return self._connections  # None or connections

        # Lazy func is not None
        # Connections may be None or not
        if time.time() > self._cache_valid_until:
            return request_data()
        # Cache time is valid
        # If there is no cache yet, request the data
        if self._connections is None:
            return request_data()
        # Return cached connections
        return self._connections



"""
Exceptions for the onboardapis package.
"""


class OnboardException(Exception):
    """
    Base exception class for on-board exceptions.
    """

    __slots__ = []


class APIConnectionError(OnboardException, ConnectionError):
    """
    Exception raised when an API connection error occurs.
    """

    __slots__ = []


class InitialConnectionError(APIConnectionError):
    """
    Exception raised only when the first try to connect to an API fails.
    """

    __slots__ = []

    def __init__(self, message: str = "Unable to connect to the API, are you connected to the on-board WI-FI?", *args):
        super(InitialConnectionError, self).__init__(message, *args)


class DataInvalidError(APIConnectionError):
    """
    Error raised when an API sends a response that could not be parsed due to a formatting error
    or when the data is missing
    """

    __slots__ = []

    def __init__(self, message: str = "The API returned an invalid response", *args):
        super(DataInvalidError, self).__init__(message, *args)



"""
Convert between different units
"""

from typing import Tuple, Union, Optional


def coordinates_to_distance(start: Tuple[float, float], end: Tuple[float, float]) -> float:
    """
    Convert the two tuples ``start`` and ``end`` of coordinates
    into the ``distance`` in meters that lies between those two positions

    Currently not implemented!

    :param start: The start coordinates
    :type start: Tuple[float, float]
    :param end: The end coordinates
    :type end: Tuple[float, float]
    :return: The distance in meters between 'start' and 'end'
    :rtype: float
    """
    raise NotImplementedError()


def coordinates_decimal_to_dms(coordinates: Tuple[float, float]) -> Tuple[Tuple[int, int, float], Tuple[int, int, float]]:
    """
    Convert the tuple ``coordinates`` of coordinates to degrees, minutes, seconds

    :param coordinates: The decimal coordinates to convert to degrees, minutes, seconds
    :type coordinates: Tuple[float, float]
    :return: The coordinates in degrees, minutes, seconds
    :rtype: Tuple[Tuple[int, int, float], Tuple[int, int, float]]
    """
    lat, lon = coordinates
    lat_deg = int(abs(lat))
    lat_min = int((abs(lat) - lat_deg) * 60)
    lat_sec = ((abs(lat) - lat_deg) * 60 - lat_min) * 60
    lon_deg = int(abs(lon))
    lon_min = int((abs(lon) - lon_deg) * 60)
    lon_sec = ((abs(lon) - lon_deg) * 60 - lon_min) * 60
    return (-lat_deg if lat < 0 else lat_deg, lat_min, lat_sec), (-lon_deg if lon < 0 else lon_deg, lon_min, lon_sec)


def coordinates_dms_to_decimal(coordinates: Tuple[Tuple[int, int, float], Tuple[int, int, float]]) -> Tuple[float, float]:
    """
    Convert the tuple ``coordinates`` of degrees, minutes, seconds to decimal degrees

    :param coordinates: The degrees, minutes, seconds coordinates to convert to decimal degrees
    :type coordinates: Tuple[Tuple[int, int, float], Tuple[int, int, float]]
    :return: The coordinates in decimal degrees
    :rtype: Tuple[float, float]
    """
    (lat_deg, lat_min, lat_sec), (lon_deg, lon_min, lon_sec) = coordinates
    lat = lat_deg + lat_min / 60 + lat_sec / 3600
    lon = lon_deg + lon_min / 60 + lon_sec / 3600
    return lat, lon


# Convert between different units


def ms_to_kmh(meters_per_second: float) -> float:
    """
    Convert meters/second into kilometers/hour

    :param meters_per_second: The value in meters/second
    :type meters_per_second: float
    :return: The result in kilometers/hour
    :rtype: float
    """
    return meters_per_second * 3.6


def kmh_to_ms(kilometers_per_hour: float) -> float:
    """
    Convert kilometers/hour into meters/second

    :param kilometers_per_hour: The value in kilometers/hour
    :type kilometers_per_hour: float
    :return: The result in meters/second
    :rtype: float
    """
    return kilometers_per_hour / 3.6


def ms_to_mph(meters_per_second) -> float:
    """
    Convert meters/second into miles/hour

    :param meters_per_second: The value in meters/second
    :type meters_per_second: float
    :return: The result in miles/hour
    :rtype: float
    """
    return meters_per_second / 0.44704


def mph_to_ms(miles_per_hour):
    """
    Convert miles/hour into meters/second

    :param miles_per_hour: The value in miles/hour
    :type miles_per_hour: float
    :return: The result in meters/second
    :rtype: float
    """
    return miles_per_hour * 0.44704


def ms_to_kn(meters_per_second):
    """
    Convert meters/second into knots (nautical miles/hour)

    :param meters_per_second: The value in meters/second
    :type meters_per_second: float
    :return: The result in knots (nautical miles/hour)
    :rtype: float
    """
    return meters_per_second * 3600 / 1852


def kn_to_ms(knots):
    """
    Convert knots (nautical miles/hour) into meters/second

    :param knots: The value in knots (nautical miles/hour)
    :type knots: float
    :return: The result in meters/second
    :rtype: float
    """
    return knots * 1852 / 3600



"""
Implementation of the german operator DB (Deutsche Bahn).
"""

import datetime
import re

from typing import Tuple, Dict, Any, List, Optional, Literal

import requests

from .. import Train, Station, ConnectingTrain, _LazyStation
from ...exceptions import DataInvalidError, APIConnectionError
from ...utils.conversions import kmh_to_ms
from ...utils.data import (
    DynamicDataConnector, JSONDataConnector, StaticDataConnector, some_or_default, ScheduledEvent, Position
)

API_BASE_URL_ICEPORTAL = "iceportal.de"
InternetStatus = Literal["NO_INFO", "NO_INTERNET", "UNSTABLE", "WEAK", "MIDDLE", "HIGH"]


class _ICEPortalStaticConnector(StaticDataConnector, JSONDataConnector):
    __slots__ = []

    def __init__(self):
        super().__init__(base_url=API_BASE_URL_ICEPORTAL)

    def refresh(self):
        self.store("bap", self._get("/bap/api/bap-service-status"))


class _ICEPortalDynamicConnector(DynamicDataConnector, JSONDataConnector):
    __slots__ = ["_connections_cache_control"]

    def __init__(self):
        super().__init__(base_url=API_BASE_URL_ICEPORTAL)
        self._connections_cache_control = {}
        """
        A cache for connections
        Connections for DB are only available shortly before the arrival
        -> Cache every already seen connection as well
        """

    def refresh(self):
        self.store("status", self._get("/api1/rs/status"))
        self.store("trip", self._get("/api1/rs/tripInfo/trip"))

    def connections(self, station_id: str) -> List[ConnectingTrain]:
        """
        Get all connections for a station

        :param station_id: The station to get the connections for
        :type station_id: str
        :return: A list of connections for the station
        :rtype: List[ConnectingTrain]
        """
        # Function to determine when to update the cache
        def cache_valid() -> bool:
            last_update = self._connections_cache_control.get(station_id, {}).get("last_update")
            if last_update is None:
                return False
            return datetime.datetime.now() < last_update + datetime.timedelta(minutes=1)

        # Let the cache expire after 1 minute
        if cache_valid():
            return self.load(f"{station_id}_connections", [])

        # Request the connections
        try:
            connections_json = self._get(f"/api1/rs/tripInfo/connection/{station_id}")
        except APIConnectionError:
            return list()

        # Process the connections
        connections = list([
            ConnectingTrain(
                train_type=connection.get('trainType', None),
                line_number=connection.get('vzn', None),
                platform=ScheduledEvent(
                    scheduled=connection.get('track', {}).get('scheduled', None),
                    actual=connection.get('track', {}).get('actual', None),
                ),
                destination=connection.get('station', {}).get('name', None),
                departure=ScheduledEvent(
                    scheduled=(
                        datetime.datetime.fromtimestamp(int(some_or_default(
                            connection.get('timetable', {}).get('scheduledDepartureTime'),
                            default=0
                        )) / 1000)
                        if some_or_default(connection.get('timetable', {}).get('scheduledDepartureTime')) is not None
                        else None
                    ),
                    actual=(
                        datetime.datetime.fromtimestamp((some_or_default(
                            connection.get('timetable', {}).get('actualDepartureTime'),
                            default=0
                        )) / 1000)
                        if some_or_default(connection.get('timetable', {}).get('actualDepartureTime')) is not None
                        else None
                    ),
                )
            )
            for connection in connections_json.get('connections', [])
        ])
        self.store(f"{station_id}_connections", connections)
        self._connections_cache_control[station_id] = {"last_update": datetime.datetime.now()}
        return connections


class ICEPortal(Train):
    """
    Wrapper for interacting with the DB ICE Portal API
    """
    __slots__ = []
    __cached_names = dict()
    """
    Cache for train names
    """

    def __init__(self):
        super().__init__()
        self._static_data = _ICEPortalStaticConnector()
        self._dynamic_data = _ICEPortalDynamicConnector()

    def now(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(
            int(some_or_default(self._dynamic_data.load("status", {}).get('serverTime', None), default=0)) / 1000
        )

    @property
    def id(self) -> str:
        return self._dynamic_data.load("status", {}).get('tzn')

    @property
    def type(self) -> str:
        return self._dynamic_data.load("trip", {}).get('trip', {}).get('trainType')

    @property
    def number(self) -> str:
        return self._dynamic_data.load("trip", {}).get('trip', {}).get('vzn')

    @property
    def stations(self) -> Dict[str, Station]:
        # Each stations connecting trains require an additional request
        # So use a LazyStation to only request the connections when needed
        return {
            stop.get('station', {}).get('evaNr'): _LazyStation(
                station_id=stop.get('station', {}).get('evaNr'),
                name=stop.get('station', {}).get('name'),
                platform=ScheduledEvent(
                    scheduled=stop.get('track', {}).get('scheduled'),
                    actual=stop.get('track', {}).get('actual')
                ),
                arrival=ScheduledEvent(
                    scheduled=datetime.datetime.fromtimestamp(
                        int(some_or_default(stop.get('timetable', {}).get('scheduledArrivalTime'), default=0)) / 1000
                    ) if some_or_default(stop.get('timetable', {}).get('scheduledArrivalTime')) is not None else None,
                    actual=datetime.datetime.fromtimestamp(
                        (some_or_default(stop.get('timetable', {}).get('actualArrivalTime'), default=0)) / 1000
                    ) if some_or_default(stop.get('timetable', {}).get('actualArrivalTime')) is not None else None,
                ),
                departure=ScheduledEvent(
                    scheduled=datetime.datetime.fromtimestamp(
                        int(some_or_default(stop.get('timetable', {}).get('scheduledDepartureTime'), default=0)) / 1000
                    ) if some_or_default(stop.get('timetable', {}).get('scheduledDepartureTime')) is not None else None,
                    actual=datetime.datetime.fromtimestamp(
                        int(some_or_default(stop.get('timetable', {}).get('actualDepartureTime'), default=0)) / 1000
                    ) if some_or_default(stop.get('timetable', {}).get('actualDepartureTime')) is not None else None,
                ),
                position=Position(
                    latitude=stop.get('station', {}).get('geocoordinates', {}).get('latitude'),
                    longitude=stop.get('station', {}).get('geocoordinates', {}).get('longitude')
                ),
                distance=stop.get('info', {}).get('distanceFromStart', 0),
                connections=None,
                lazy_func=lambda: self._dynamic_data.connections(stop.get('station', {}).get('evaNr')),
            )
            for stop in self._dynamic_data.load("trip", {}).get('trip', {}).get('stops', [])
        }

    @property
    def origin(self) -> Station:
        return super(ICEPortal, self).origin

    @property
    def current_station(self) -> Station:
        # Get the current station id
        stop_info = self._dynamic_data.load("trip", {}).get('trip', {}).get('stopInfo', {})
        station_id = some_or_default(stop_info.get('actualNext'))
        # Get the station from the stations dict
        try:
            return self.stations[station_id]
        except AttributeError as e:
            raise DataInvalidError("No current station found") from e

    @property
    def destination(self) -> Station:
        return super(ICEPortal, self).destination

    @property
    def speed(self) -> float:
        return kmh_to_ms(self._dynamic_data.load("status", {}).get('speed', 0))

    @property
    def distance(self) -> float:
        return (
            self._dynamic_data.load("trip", {}).get('trip', {}).get('actualPosition', 0)
            + self._dynamic_data.load("trip", {}).get('trip', {}).get('distanceFromLastStop', 0)
        )

    @property
    def position(self) -> Position:
        return Position(
            self._dynamic_data.load("status", {}).get('latitude'),
            self._dynamic_data.load("status", {}).get('longitude')
        )

    @property
    def delay(self) -> float:
        return super(ICEPortal, self).delay

    @property
    def name(self) -> Optional[str]:
        """
        Get the name of the train.

        Most of the DB ICE trains have names.
        Names are not available through the API, instead a public list of names will be used.

        :return: The name of the train
        """
        # Check if train names are already cached
        if len(self.__cached_names) == 0:
            # Request the list of train names
            data = requests.get(
                "https://felix-zenk.github.io/projects/onboardapis/res/trains/germany/db/names.json",
                headers={"user-agent": "python-onboardapis", "accept": "application/json"}
            ).json()
            # Add the names to the cache
            self.__cached_names.update(data.get('names', {}))
            # Also use unverified train names, because there are not many verified names at this point
            self.__cached_names.update(data.get('unverified_names', {}))

        # Get the name from the cache
        match = re.search(r'\d+', self.id)
        if match is None:
            return None

        return self.__cached_names.get(match.group(0).lstrip('0'), None)

    def all_delay_reasons(self) -> Dict[str, Optional[List[str]]]:
        """
        Get all delay reasons for the current trip

        :return: A dictionary of delay reasons with the station id as the key
        :rtype: Dict[str, Optional[List[str]]]
        """
        return {
            stop.get('station', {}).get('evaNr', None): list([
                some_or_default(reason.get('text', None)) for reason in some_or_default(stop.get('delayReasons'), [])
            ])
            for stop in self._dynamic_data.load("trip", {}).get('trip', {}).get('stops', [])
        }

    def delay_reasons(self) -> Optional[List[str]]:
        """
        Get the delay reason for the current station

        :return: The delay reason
        :rtype: Optional[List[str]]
        """
        return self.all_delay_reasons().get(self.current_station.id, None)

    def has_bap(self) -> bool:
        """
        Returns True if the train has a BAP module installed and active

        bap = 'Bestellen am Platz' is a service that allows passengers to order food and drinks right to their seat

        :return:
        """
        # bap is a service exclusive to first class
        if self.wagon_class() != "FIRST":
            return False
        # Check if the module is installed
        if str(self._dynamic_data.load("status", {}).get('bapInstalled', False)).lower() != "true":
            return False
        # Check if the module is active
        if str(self._static_data.load("bap", {}).get('status', False)).lower() == "true":
            return True
        return False

    def wagon_class(self) -> Optional[Literal["FIRST", "SECOND"]]:
        """
        Get the wagon class of the wagon you are currently in

        :return: The wagon class
        :rtype: Literal["FIRST", "SECOND"]
        """
        return some_or_default(self._dynamic_data.load("status", {}).get('wagonClass'))

    def internet_connection(self) -> Tuple[
        Optional[InternetStatus], Optional[InternetStatus], Optional[datetime.timedelta]
    ]:
        """
        Returns the internet connection status of the train,

        the next internet connection status,

        and the time until the change occurs.

        Be aware that some or all values can be None.

        :return: The tuple (current, next, time_remaining)
        :rtype: Tuple[InternetStatus, InternetStatus, datetime.timedelta]
        """
        return (
            # Current state
            some_or_default(self._dynamic_data.load("status", {}).get('connectivity', {}).get('currentState')),
            # Next state
            some_or_default(self._dynamic_data.load("status", {}).get('connectivity', {}).get('nextState')),
            # Remaining time
            datetime.timedelta(
                seconds=int(self._dynamic_data.load("status", {}).get('connectivity', {}).get('remainingTimeSeconds'))
            )
            if some_or_default(
                self._dynamic_data.load("status", {}).get('connectivity', {}).get('remainingTimeSeconds', '')
            ) is not None else None
        )



"""
This module contains everything that has to do with data and data management
"""

import abc
import json
import threading
import time
from os import PathLike

import requests

from typing import Any, Optional, TypeVar, Generic, ItemsView, Union

from requests import Response, RequestException

from .conversions import coordinates_decimal_to_dms
from ..exceptions import DataInvalidError, APIConnectionError, InitialConnectionError
from .. import __version__


def some_or_default(data: Optional[Any], default: Optional[Any] = None) -> Optional[Any]:
    """
    Return ``data`` if there is actually some content in data, else return ``default``.

    Useful when data such as "" or b'' should also be treated as empty.

    :param data: The data to test
    :type data: Any
    :param default: The default value to return if no data is present
    :type default: Any
    :return: The data if present, else the default
    :rtype: Optional[Any]
    """
    if data is None:
        return default
    if isinstance(data, str) and data == "":
        return default
    if isinstance(data, bytes) and data == b"":
        return default
    # if isinstance(data, Sized) and len(data) == 0:
    #     return default
    return data


T = TypeVar('T')
"""
A type variable for generic functions
"""


class ScheduledEvent(Generic[T]):
    """
    Something that is scheduled and can happen as ``scheduled``,
    but can also happen different from the expected and actually happens as ``actual``
    """

    __slots__ = ["scheduled", "actual"]

    def __init__(self, scheduled: Optional[T] = None, actual: Optional[T] = None):
        """
        Initialize a new :class:`ScheduledEvent`

        :param scheduled: The value that should happen
        :type scheduled: Optional[T]
        :param actual: The value that actually happens
        :type actual: Optional[T]
        """
        self.scheduled = scheduled
        self.actual = actual

    def __repr__(self):
        if self.actual is None:
            return f"<{self.__class__.__name__} {self.scheduled}>"
        return f"<{self.__class__.__name__} {self.actual}>"

    def __str__(self):
        if self.actual is None:
            return f"{self.scheduled}"
        return f"{self.actual}"


class Position(object):
    """
    A position requires at least a latitude and longitude,
    but can also provide data on altitude and the current compass heading.
    """

    __slots__ = ["_latitude", "_longitude", "_altitude", "_heading"]

    def __init__(self, latitude: float, longitude: float, altitude: float = None, bearing: float = None):
        """
        Initialize a new :class:`Position`.

        :param latitude: The latitude in decimal degrees
        :type latitude: float
        :param longitude: The longitude in decimal degrees
        :type longitude: float
        :param altitude: The altitude in meters
        :type altitude: float
        :param bearing: The compass heading in degrees
        :type bearing: float
        """
        self._latitude = latitude
        self._longitude = longitude
        self._altitude = altitude
        self._heading = bearing

    def __str__(self) -> str:
        (lat_deg, lat_min, lat_sec), (lon_deg, lon_min, lon_sec) = coordinates_decimal_to_dms(
            (self.latitude, self.longitude)
        )
        coordinates = (
            f"{abs(lat_deg)}°{lat_min}'{lat_sec:.3f}\"{'N' if lat_deg >= 0 else 'S'} "
            f"{abs(lon_deg)}°{lon_min}'{lon_sec:.3f}\"{'E' if lon_deg >= 0 else 'W'}"
        )
        return coordinates

    def __getitem__(self, item):
        return list([self.latitude, self.longitude])[item]

    @property
    def latitude(self) -> float:
        """
        The latitude in decimal degrees.

        :return: The latitude
        :rtype: float
        """
        return self._latitude

    @property
    def longitude(self) -> float:
        """
        The longitude in decimal degrees.

        :return: The longitude
        :rtype: float
        """
        return self._longitude

    @property
    def altitude(self) -> float:
        """
        The altitude in meters.

        :return: The altitude
        :rtype: float
        """
        return self._altitude

    @property
    def heading(self) -> float:
        """
        The compass heading in degrees.

        0 is north, 90 is east, 180 is south, 270 is west.

        :return: The heading
        :rtype: float
        """
        return self._heading

    def calculate_distance(self, other: "Position") -> float:
        """
        Calculate the distance (in meters) between this position and another position.

        :param other: The other position
        :return: The distance in meters
        """
        raise NotImplementedError()
        # return coordinates_to_distance((self.latitude, self.longitude), (other.latitude, other.longitude))


class DataStorage:
    """
    A storage class that can be used to store data and retrieve it later.
    """

    def get(self, key: str) -> Any:
        """
        Get the value of the given key out of the storage.

        Raises an :class:`AttributeError` if the key is not present.

        :param key: The key to get the value of
        :type key: str
        :return: The stored value
        :rtype: Any
        """
        return getattr(self, key)

    def set(self, key: str, value: Any) -> None:
        """
        Set the value of the given key in the storage.

        Overwrites existing keys.

        :param key: The key to set the value of
        :type key: str
        :param value: The value to store
        :type value: Any
        :return: Nothing
        :rtype: None
        """
        setattr(self, key, value)

    def remove(self, key: str) -> None:
        """
        Remove the given key and its value from the storage.

        :param key: The key to remove
        :type key: str
        :return: Nothing
        :rtype: None
        """
        delattr(self, key)

    def items(self) -> ItemsView[str, Any]:
        """
        Get all keys and values from the storage.

        :return: A view of all keys and values
        :rtype: ItemsView[str, Any]
        """
        return vars(self).items()


class DataConnector(metaclass=abc.ABCMeta):
    """
    A class for retrieving data from an API
    """

    __slots__ = ["base_url", "_session", "_cache", "_verify"]

    def __init__(self, base_url: str, *, verify: bool = True) -> None:
        """
        Initialize a new :class:`DataConnector`

        :param base_url: The base url of the server to connect to
        :type base_url: str
        :param verify: Whether to verify the TLS certificate of the server
        :type verify: bool
        """
        self.base_url = base_url
        self._session = requests.Session()
        self._cache = DataStorage()
        self._verify = verify

    def __del__(self):
        self._session.close()

    def _get(self, endpoint: str, *args: Any, **kwargs: Any) -> Response:
        """
        Request data from the server

        :param endpoint: The endpoint to request data from
        :type endpoint: str
        :param args: args to pass to the request
        :type args: Any
        :param kwargs: kwargs to pass to the request
        :type kwargs: Any
        :return: The response from the server
        :rtype: Response
        """
        # Overwrite some kwargs
        headers = kwargs.get("headers", {})
        headers.update({
            "user-agent": f"python-onboardapis/{__version__}"
        })
        kwargs.update({
            "headers": headers,
            "timeout": 1,
            "verify": kwargs.get("verify", self._verify)
        })
        try:
            return self._session.get(f"https://{self.base_url}{endpoint}", *args, **kwargs)
        except RequestException as e:
            raise APIConnectionError() from e

    def load(self, key: str, __default: Any = None) -> Any:
        """
        Load data from the cache

        :param key: The key to load
        :param __default: The default value to return if the key is not present
        :return: The data if present, else the default
        """
        try:
            return self._cache.get(key)
        except AttributeError:
            return __default

    def store(self, key: str, value: Any) -> None:
        """
        Store data in the cache

        :param key: The key the data should be stored under
        :param value: The data to store
        :return: Nothing
        """
        self._cache.set(key, value)

    def export(self, path: Union[str, PathLike]) -> None:
        """
        Export the cache to a file

        :param path: The path to export to
        :type path: Union[str, PathLike]
        :return: Nothing
        :rtype: None
        """
        data = {}
        for key, value in self._cache.items():
            data[key] = value

        with open(path, "w") as f:
            json.dump(data, f)

    @abc.abstractmethod
    def refresh(self) -> None:  # pragma: no cover
        """
        Method that collects data from the server and stores it in the cache

        :return: Nothing
        """
        pass


class StaticDataConnector(DataConnector, metaclass=abc.ABCMeta):
    """
    A :class:`DataConnector` for data never changes and therefore has to be requested only once
    """

    __slots__ = []

    def __init__(self, base_url: str, *args, **kwargs):
        super().__init__(base_url, *args, **kwargs)


class DynamicDataConnector(DataConnector, metaclass=abc.ABCMeta):
    """
    A :class:`DataConnector` for data that changes frequently and therefore has to be requested continuously
    """

    __slots__ = ["_runner", "_running", "_initialized", "optimize"]

    def __init__(self, base_url: str, *args, **kwargs):
        super().__init__(base_url, *args, **kwargs)
        self._runner = threading.Thread(
            target=self._run, name=f"DynamicDataConnector-Runner for '{self.base_url}'", daemon=True
        )
        self._running = False
        self._initialized = False
        self.optimize = True

    @property
    def connected(self) -> bool:
        """
        Check whether the connector is connected to the server
        """
        return self._initialized and self._running

    def _run(self) -> None:
        """
        The main loop that will run in a separate thread

        :return: Nothing
        :rtype: None
        """
        # tps is the amount of checks per second for a thread join while waiting for the next refresh
        # During a refresh it does not check for a thread join
        tps = 20  # 20 ticks per second -> check for thread join every 0.05 seconds, refresh the data each second

        counter = 0
        while self._running:
            # If the counter is not 0, just wait and check for a thread join (self._running == False)
            if counter != 0:
                time.sleep(1 / tps)
                counter = (counter + 1) % tps
                continue

            # The target time for when to perform the next refresh after this one
            target = time.time_ns() + int(1e9)

            # Perform the actual refresh
            try:
                self.refresh()
            except APIConnectionError as e:
                self._running = False
                raise e

            # Signal that the data has been refreshed at least once (for thread synchronization)
            if not self._initialized:
                self._initialized = True

            # If optimize is enabled, measure the time it took to refresh the data
            # and calculate the remaining ticks until target time
            if self.optimize:
                counter = (tps - int(max(0.0, (target - time.time_ns()) / 1e9) * tps)) % tps
            else:
                counter = (counter + 1) % tps

    def start(self) -> None:
        """
        Start requesting data
        """
        self._running = True
        self._runner.start()
        while self._runner.is_alive() and not self._initialized:  # pragma: no cover
            # Wait until the new thread has initialized (received data at least once)
            pass

        if not self._runner.is_alive():
            # If the thread is not alive, something went wrong
            raise InitialConnectionError("Failed to connect to the server")

    def stop(self):
        """
        Stop requesting data and shut down the separate thread
        """
        self._running = False
        if self._runner.is_alive():
            self._runner.join()

    def reset(self):
        """
        Reset the thread and the cache so that they can be reused with ``start()``
        """
        self.stop()
        self._runner = threading.Thread(
            target=self._run, name=f"DynamicDataConnector-Runner for '{self.base_url}'", daemon=True
        )
        self._session = requests.Session()
        self._cache = DataStorage()
        self._running = False
        self._initialized = False


class JSONDataConnector(DataConnector, metaclass=abc.ABCMeta):
    """
    A :class:`DataConnector` that automatically parses the response to a json object
    """

    __slots__ = []

    def _get(self, endpoint: str, *args: Any, **kwargs: Any) -> dict:
        headers = kwargs.get("headers", {})
        headers.update({
            "accept": "application/json",
        })
        kwargs["headers"] = headers
        try:
            return super(JSONDataConnector, self)._get(endpoint, *args, **kwargs).json()
        except json.JSONDecodeError as e:
            raise DataInvalidError() from e



from setuptools import setup, find_packages

from onboardapis import __project_name__, __description__, __version__, __author__, __email__, __license__, __url__

with open('README.md', mode='r', encoding='utf-8') as f:
    readme = f.read()

setup(
    name=__project_name__,
    version=__version__,
    packages=find_packages(".", exclude=["tests", "tests.*", "examples", "examples.*"]),
    url=f"https://felix-zenk.github.io/projects/{__project_name__}",
    license=__license__,
    author=__author__,
    author_email=__email__,
    description=__description__,
    long_description=readme,
    long_description_content_type='text/markdown',
    project_urls={
        'Bug Tracker': f"{__url__}/issues",
        'Source': __url__,
        'Documentation': f"https://{__project_name__}.readthedocs.io/en/latest",
    },
    python_requires='>=3.6',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Utilities',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)




"""
Package metadata
"""

__project_name__ = 'onboardapis'
__description__ = 'A pure Python wrapper for the on-board APIs of many different transportation providers'
__version__ = '1.2.2'
__author__ = 'Felix Zenk'
__email__ = 'felix.zenk@web.de'
__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2022 Felix Zenk'
__url__ = 'https://github.com/felix-zenk/onboardapis'


class Vehicle(object):
    """
    The base class for all vehicles.
    """

    __slots__ = []
